/*
 * Test case for recursive mutex issue
 */

#define _GNU_SOURCE 1
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>

static pthread_mutex_t test_lock = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static pthread_mutex_t test_lock2 = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static pthread_t test_thread; 
static int lock_count = 0;
static int test_errors = 0;

/*! \brief This is a regression test for recursive mutexes.
   test_for_thread_safety() will return 0 if recursive mutex locks are
   working properly, and non-zero if they are not working properly. */
static void *test_thread_body(void *data) 
{ 
	int res;
	res = pthread_mutex_lock(&test_lock);
	fprintf(stderr, "LOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count += 10;
	if (lock_count != 10) {
		fprintf(stderr, "%s:%d lock_count=%d != 10\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	res = pthread_mutex_lock(&test_lock);
	fprintf(stderr, "LOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count += 10;
	if (lock_count != 20) {
		fprintf(stderr, "%s:%d lock_count=%d != 20\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	pthread_mutex_lock(&test_lock2);
	res = pthread_mutex_unlock(&test_lock);
	fprintf(stderr, "UNLOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count -= 10;
	if (lock_count != 10) {
		fprintf(stderr, "%s:%d lock_count=%d != 10\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	res = pthread_mutex_unlock(&test_lock);
	fprintf(stderr, "UNLOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count -= 10;
	pthread_mutex_unlock(&test_lock2);
	if (lock_count != 0) {
		fprintf(stderr, "%s:%d lock_count=%d != 0\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	return NULL;
} 

int test_for_thread_safety(void)
{ 
	int res;
	pthread_mutex_lock(&test_lock2);
	res = pthread_mutex_lock(&test_lock);
	fprintf(stderr, "LOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count += 1;
	res = pthread_mutex_lock(&test_lock);
	fprintf(stderr, "LOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count += 1;
	pthread_create(&test_thread, NULL, test_thread_body, NULL); 
	usleep(100);
	if (lock_count != 2) {
		fprintf(stderr, "%s:%d lock_count=%d != 2\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	res = pthread_mutex_unlock(&test_lock);
	fprintf(stderr, "UNLOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count -= 1;
	usleep(100); 
	if (lock_count != 1) {
		fprintf(stderr, "%s:%d lock_count=%d != 1\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	res = pthread_mutex_unlock(&test_lock);
	fprintf(stderr, "UNLOCK %s:%d res=%d\n", __PRETTY_FUNCTION__, __LINE__, res);
	lock_count -= 1;
	if (lock_count != 0) {
		fprintf(stderr, "%s:%d lock_count=%d != 0\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	pthread_mutex_unlock(&test_lock2);
	usleep(100);
	if (lock_count != 0) {
		fprintf(stderr, "%s:%d lock_count=%d != 0\n", __PRETTY_FUNCTION__, __LINE__, lock_count);
		test_errors++;
	}
	pthread_join(test_thread, NULL);
	return(test_errors);          /* return 0 on success. */
}

int main(int argc, char *argv[])
{
	if (test_for_thread_safety()) {
		fprintf(stderr, "Mutexes do not behave as recursive mutexes.\n");
	} else {
		fprintf(stderr, "A-okay\n");
	}
	return 0;
}

